#include <stdarg.h>
#include <stddef.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/config.h>
#include <linux/pci.h>
#include <linux/proc_fs.h>
#include <linux/cobalt_i2c.h>
#include <asm/io.h>

#if defined(CONFIG_COBALT_GEN_III) || defined(CONFIG_COBALT_GEN_V)

static unsigned int max_temp_id = 0;

char *cobalt_temp_read(unsigned int);
static char *degrees(int);

void __init cobalt_temp_init(void)
{
	int type = cobalt_i2c_type();

	switch (type) {
	case COBALT_I2C_TYPE_GEN_V:
		max_temp_id = 3; /* monterey has 4 temp sensors */
	case COBALT_I2C_TYPE_GEN_III:
		printk("Cobalt temperature sensor v1.3 enabled\n");
		break;
	default:
		printk("Cobalt temperature sensor not detected\n");
	}
}

#define LM77_TEMP		0x0
#define LM77_STATUS		0x1
#define LM77_HYST		0x2
#define LM77_CRIT		0x3
#define LM77_LOW		0x4
#define LM77_HIGH		0x5

#ifndef min
#define min(a,b)  ((a) < (b) ? (a) : (b))
#endif

/* Sensors on XTR:
 * 0 - CPU0, 1 - CPU1 (U31), 2 - behind RAM (U60), 3 - center (U59) 
 */
char *cobalt_temp_read(unsigned int sensor)
{
	int tmp;
	int val = 0;
	int tries = 2;
	static int last_val = 0;

	sensor = min(sensor, max_temp_id);

	/* sometimes it reads as zero... try again */
	while (tries--) {
		/* LM77 returns the bytes backwards - <shrug> */
		/* address = base + deviceid + 1 for read */
        	tmp = cobalt_i2c_read_word(COBALT_I2C_DEV_TEMP + (sensor<<1) + 1, LM77_TEMP);
		if (tmp < 0) {
			/* read failed, return the last known value */
			return degrees(last_val);
		}

        	val = (tmp<<8 & 0xff00) + (tmp>>8 & 0x00ff);
		if (val) {
			last_val = val;
			break;
		}
	}
	return degrees(val);
}

/* build a string - measured in degrees C */
static char *degrees(int val)
{
	static char temp[16] = "unknown";

	sprintf(temp, "%d", val>>4);
	if (val & 0x0008) {
		sprintf(temp+strlen(temp), ".5");
	}
	return temp;
}
#endif /* CONFIG_COBALT_GEN_III */
