/* 
 * Cobalt kernel WDT timer driver
 * Tim Hockin <thockin@cobaltnet.com>
 * RaQ 5 --> <asun@cobalt.com>
 * Copyright (c)1999-2000, Cobalt Networks
 *
 * this is pretty ugly.
 */
#include <linux/autoconf.h>

#if 0
#if CONFIG_MODVERSIONS == 1 && ! defined(MODVERSIONS)
#define MODVERSIONS
#endif

#ifdef MODVERSIONS
#include <linux/modversions.h>
#endif
#endif

#include <linux/module.h>
#include <stdarg.h>
#include <stddef.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/config.h>
#include <linux/pci.h>
#include <linux/delay.h>

#if defined(CONFIG_COBALT_GEN_III) || defined(CONFIG_COBALT_GEN_V)
#include <linux/cobalt_wdt.h>
#include <linux/pc87317.h>
#include <asm/io.h>

#define DOGB		0x20
#define ALI_7101_WDT	0x92
#define ALI_WDT_ARM	0x01

#define SUPERIO_TIMEOUT  (0x01)    /* 1 minute */
#define RAQ5_WDT_TIMEOUT (HZ << 3)

#define	WDT_UNINITIALIZED	0
#define	WDT_PIC			1
#define	WDT_ALIM1543		2
#define	WDT_PC87317		3

#define WDT_FLAG_BOOTLOADER	(1 << 16)

#define WDT_TYPE(a)		((a) & 0xffff)
#define WDT_IGNORE(a)		((WDT_TYPE(a) == WDT_UNINITIALIZED) || \
				 ((a) & WDT_FLAG_BOOTLOADER))

static int cobalt_wdt_disable_count = 0;
static struct timer_list cobalt_wdt_timer;
static unsigned long long tsc_per_wdt; 
static struct pci_dev *cobalt_pmu;
static u16 superio_addr;
static unsigned long wdt_timeout;
static unsigned int wdt = WDT_UNINITIALIZED;
static spinlock_t wdt_lock = SPIN_LOCK_UNLOCKED;

static inline unsigned long long cobalt_wdt_gettsc(void) 
{
	unsigned long long ret;
	asm volatile ("rdtsc" : "=A" (ret));
	return ret;
}


/* set up the Nat. Semi SuperI/O for Monterey. it has a minimum
 * of a 1 minute timeout. */
static unsigned long __init pc87317_setup(void)
{
	/* superi/o -- select pm logical device and get base address */
	outb(SUPERIO_LOGICAL_DEV, SUPERIO_INDEX_PORT);
	outb(SUPERIO_DEV_PM, SUPERIO_DATA_PORT);
	outb(SUPERIO_ADDR_HIGH, SUPERIO_INDEX_PORT);
	superio_addr = inb(SUPERIO_DATA_PORT) << 8;
	outb(SUPERIO_ADDR_LOW, SUPERIO_INDEX_PORT);
	superio_addr |= inb(SUPERIO_DATA_PORT); 
	if (superio_addr)
		wdt = WDT_PC87317;
	return RAQ5_WDT_TIMEOUT;
}


static unsigned long  __init alim7101_setup(void)
{
	struct pci_dev *south;
	char tmp;

	/* PMU (1543 ver A1-E) has a built-in WDT.  Set it to 1 second */
	pci_write_config_byte(cobalt_pmu, ALI_7101_WDT, 0x02);

	/* reversion # is here - must match ???1001?(b) or use PIC WDT */
	/* why it is called 1543, but DevId is 1533 I'll never know */
	south = pci_find_device(PCI_VENDOR_ID_AL, PCI_DEVICE_ID_AL_M1533, NULL);
	if (!south) {
		printk("cobalt_wdt_init: can't find 1543 device!\n");
		wdt = WDT_PIC;
	} else {
		pci_read_config_byte(south, 0x5e, &tmp);
		wdt = ((tmp & 0x1e) != 0x12) ? WDT_PIC : WDT_ALIM1543;
	}

	if (WDT_TYPE(wdt) == WDT_ALIM1543) {
		/* set DOGB GPIO pin to OUTPUT - JIC */
		pci_read_config_byte(cobalt_pmu, 0x7d, &tmp);
		pci_write_config_byte(cobalt_pmu, 0x7d, tmp | DOGB);
	}
	return WDT_TIMEOUT;
}

void __init cobalt_wdt_init(void)
{
	unsigned long long start, stop;

	/* gen v uses a nat semi superi/o for its watchdog. it's connected
	 * to the OSB4. */
	if (pci_find_device(PCI_VENDOR_ID_SERVERWORKS, 
			    PCI_DEVICE_ID_SERVERWORKS_OSB4, NULL)) {
		wdt_timeout = pc87317_setup();
	} else if ((cobalt_pmu = pci_find_device(PCI_VENDOR_ID_AL, 
						PCI_DEVICE_ID_AL_M7101, 
						NULL))) {
		wdt_timeout = alim7101_setup();
	} else {
		printk("cobalt_wdt_init: no PMU found!\n");
		return;
	}

	start = cobalt_wdt_gettsc();
	udelay(100);
	stop = cobalt_wdt_gettsc();

	/*
	 * (int) (stop - start) * 10 == tsc per msec
	 * 1000 / HZ == msec per tick
	 * wdt_timeout == ticks per watchdog rearm
	 */
	tsc_per_wdt = (int) (stop - start) * 10 * (1000 * wdt_timeout / HZ);

#ifdef CONFIG_COBALT_BOOTLOADER
	wdt |= WDT_FLAG_BOOTLOADER;
#else
	/* set the timer going */
	init_timer(&cobalt_wdt_timer);
	cobalt_wdt_timer.function = cobalt_wdt_refresh;
	cobalt_wdt_timer.data = 1;
	cobalt_wdt_timer.expires = jiffies + wdt_timeout;
	add_timer(&cobalt_wdt_timer);

	/* the first timer tick will set it going */
	printk("Cobalt watchdog v1.4 enabled");
	if (WDT_TYPE(wdt) == WDT_PIC) 
		printk(" (old revision board: using PIC controller)");
	printk("\n");
#endif
}

static void wdt_reset(void)
{
	unsigned long flags;
	char tmp;

	if (WDT_TYPE(wdt) == WDT_UNINITIALIZED)
		return;

        spin_lock_irqsave(&wdt_lock, flags);
        if (WDT_TYPE(wdt) == WDT_PC87317) {
                outb(SUPERIO_WDT_DEV, superio_addr);
                outb(SUPERIO_TIMEOUT, superio_addr + 1);
        } else {
                /* read the value, disable (reset) WDT, enable WDT */
                pci_read_config_byte(cobalt_pmu, ALI_7101_WDT, &tmp);
                pci_write_config_byte(cobalt_pmu, ALI_7101_WDT,
                                      (tmp & ~ALI_WDT_ARM));
                pci_write_config_byte(cobalt_pmu, ALI_7101_WDT,
                                      (tmp | ALI_WDT_ARM));
                if (WDT_TYPE(wdt) == WDT_PIC) {
                        /* transition GPIO 5 (DOGB) to arm/clear timer */
                        pci_read_config_byte(cobalt_pmu, 0x7e, &tmp);
                        pci_write_config_byte(cobalt_pmu, 0x7e, tmp ^ DOGB);
                }
        }
        spin_unlock_irqrestore(&wdt_lock, flags);
}

void cobalt_wdt_refresh(unsigned long refresh_timer)
{
	if (WDT_IGNORE(wdt))
		return;

	cobalt_wdt_cleardog();

	/* re-arm the timer - never called from bootloader */
	if (refresh_timer) 
		mod_timer(&cobalt_wdt_timer, jiffies + wdt_timeout);
}

EXPORT_SYMBOL(cobalt_wdt_cleardog);

void cobalt_wdt_cleardog()
{
	static unsigned long long last_tsc = 0;
	unsigned long flags;

	if (WDT_IGNORE(wdt))
		return;

	if (cobalt_wdt_disable_count > 0)
		return;

	if ((int) (cobalt_wdt_gettsc() - last_tsc) < tsc_per_wdt)
	    return;

	wdt_reset();

        spin_lock_irqsave(&wdt_lock, flags);
        last_tsc = cobalt_wdt_gettsc();
        spin_unlock_irqrestore(&wdt_lock, flags);
}

/* this is called from machine_restart. it should not be used on
 * gen v machines. */
void cobalt_wdt_trigger_reboot(void)
{
	if (WDT_TYPE(wdt) != WDT_PIC && WDT_TYPE(wdt) != WDT_ALIM1543) {
		printk("cobalt_wdt_trigger_reboot: wdt = 0x%x\n", wdt);
		printk("cobalt_wdt_trigger_reboot: reboot not possible\n");
		return;
	}

#if !defined(CONFIG_X86_REMOTE_DEBUG) && !defined(CONFIG_COBALT_BOOTLOADER)
	/* stop feeding it */
	del_timer(&cobalt_wdt_timer);
#endif
	/* set it to trigger */
	wdt_reset();

	/* don't do anything else to it */
	wdt = WDT_UNINITIALIZED;
}

EXPORT_SYMBOL(cobalt_wdt_disable);

void cobalt_wdt_disable(void)
{
	char tmp;
	unsigned long flags;

	if (WDT_IGNORE(wdt))
		return;

	if (WDT_TYPE(wdt) == WDT_PIC) {
		printk("cobalt_wdt_disable: in PIC mode - cannot disable\n");
		return;
	}

	spin_lock_irqsave(&wdt_lock, flags);
	if (cobalt_wdt_disable_count++ == 0) {
	    del_timer(&cobalt_wdt_timer);

	    if (WDT_TYPE(wdt) == WDT_PC87317) {
		    outb(SUPERIO_WDT_DEV, superio_addr);
		    outb(0x0, superio_addr + 1);
	    } else if (WDT_TYPE(wdt) == WDT_ALIM1543) {
		    /* read the value, disable (reset) WDT */
		    pci_read_config_byte(cobalt_pmu, ALI_7101_WDT, &tmp);
		    pci_write_config_byte(cobalt_pmu, ALI_7101_WDT,
					  (tmp & ~ALI_WDT_ARM));
	    }
	}
	spin_unlock_irqrestore(&wdt_lock, flags);
}

EXPORT_SYMBOL(cobalt_wdt_reenable);

void cobalt_wdt_reenable(void)
{
	unsigned long flags;

	if (WDT_IGNORE(wdt))
	    return;

	spin_lock_irqsave(&wdt_lock, flags);
	if (--cobalt_wdt_disable_count == 0) {

	    if (cobalt_wdt_disable_count < 0) {
		printk("cobalt_wdt_reenable: too many enables\n");
		cobalt_wdt_disable_count = 0;
	    }

	    cobalt_wdt_refresh(1);
	}
	spin_unlock_irqrestore(&wdt_lock, flags);
}

#endif /* CONFIG_COBALT_GEN_III || CONFIG_COBALT_GEN_V */
