#ifndef _NET_H_SYMBA
#define _NET_H_SYMBA

/* transmit status bit definitions */
#define TX_STATUS_TXOK         (1<<13)     /* success */
#define TX_STATUS_TDLC         (1<<12)     /* dropped for late colls */
#define TX_STATUS_TCXSDFR      (1<<11)     /* excessive deferral */
#define TX_STATUS_TDEC         (1<<10)     /* excessive collisions */
#define TX_STATUS_TAUR         (1<<9)      /* abort on underrun/"jumbo" */
#define TX_STATUS_PDFRD        (1<<8)      /* packet deferred */
#define TX_STATUS_BCAST        (1<<7)      /* broadcast ok */
#define TX_STATUS_MCAST        (1<<6)      /* multicast ok */
#define TX_STATUS_CRCERR       (1<<5)      /* CRC error */
#define TX_STATUS_LC           (1<<4)      /* late collision */
#define TX_STATUS_CCNT_MASK    0xf         /* collision count */

#define T_TXOK      (1<<13)
#define T_TDLC      (1<<12)
#define T_TCXSDFR   (1<<11)
#define T_TDEC      (1<<10)
#define T_TAUR      (1<<9)
#define T_PDFRD     (1<<8)
#define T_BCAST     (1<<7)
#define T_MCAST     (1<<6)
#de8 bits) */
#define T1CH		(5*RS)		/* Timer 1 counter (high 8 bits) */
#define T1LL		(6*RS)		/* Timer 1 latch (low 8 bits) */
#define T1LH		(7*RS)		/* Timer 1 latch (high 8 bits) */
#define T2CL		(8*RS)		/* Timer 2 ctr/latch (low 8 bits) */
#define T2CH		(9*RS)		/* Timer 2 counter (high 8 bits) */
#define SR		(10*RS)		/* Shift register */
#define ACR		(11*RS)		/* Auxiliary control register */
#define PCR		(12*RS)		/* Peripheral control register */
#define IFR		(13*RS)		/* Interrupt flag register */
#define IER		(14*RS)		/* Interrupt enable register */
#define ANH		(15*RS)		/* A-side data, no handshake */

/* Bits in B data register: all active low */
#define TREQ		0x08		/* Transfer request (input) */
#define TACK		0x10		/* Transfer acknowledge (output) */
#define TIP		0x20		/* Transfer in progress (output) */

/* Bits in ACR */
#define SR_CTRL		0x1c		/* Shift register control bits */
#define SR_EXT		0x0c		/* Shift on external clock */
#define SR_OUT		0x10		/* Shift out if 1 */

/* Bits in IFR and IER */
#define IER_SET		0x80		/* set bits in IER */
#define IER_CLR		0		/* clear bits in IER */
#define SR_INT		0x04		/* Shift register full/empty */

static enum cuda_state {
    idle,
    sent_first_byte,
    sending,
    reading,
    read_done,
    awaiting_reply
} cuda_state;

static struct adb_request *current_req;
static struct adb_request *last_req;
static unsigned char cuda_rbuf[16];
static unsigned char *reply_ptr;
static int reading_reply;
static int data_index;
static struct device_node *vias;
static int cuda_fully_inited = 0;

static int init_via(void);
static void cuda_start(void);
static void via_interrupt(int irq, void *arg, struct pt_regs *regs);
static void cuda_input(unsigned char *buf, int nb, struct pt_regs *regs);
static int cuda_adb_send_request(struct adb_request *req, int sync);
static int cuda_adb_autopoll(int devs);
static int cuda_adb_reset_bus(void);
static int cuda_send_request(struct adb_request *req);


static struct adb_controller	cuda_controller = {
	ADB_VIACUDA,
	cuda_adb_send_request,
	cuda_adb_autopoll,
	cuda_adb_reset_bus,
	cuda_poll
};

__openfirmware

void
find_via_cuda()
{
    vias = find_devices("via-cuda");
    if (vias == 0)
	return;
    if (vias->next != 0)
	printk(KERN_WARNING "Warning: only using 1st via-cuda\n");

#if 0
    { int i;

    printk("via_cuda_init: node = %p, addrs =", vias->node);
    for (i = 0; i < vias->n_addrs; ++i)
	printk(" %x(%x)", vias->addrs[i].address, vias->addrs[i].size);
    printk(", intrs =");
    for (i = 0; i < vias->n_intrs; ++i)
	printk(" %x", vias->intrs[i].line);
    printk("\n"); }
#endif

    if (vias->n_addrs != 1 || vias->n_intrs != 1) {
	printk(KERN_ERR "via-cuda: expecting 1 address (%d) and 1 interrupt (%d)\n",
	       vias->n_addrs, vias->n_intrs);
	if (vias->n_addrs < 1 || vias->n_intrs < 1)
	    return;
    }
    via = (volatile unsigned char *) ioremap(vias->addrs->address, 0x2000);

    cuda_state = idle;

    if (!init_via()) {
	printk(KERN_ERR "init_via failed\n");
	via = NULL;
    }

   adb_controller = &cuda_controller;
}

void
viaTransmit branch select register */
#define TX_BRANCH_SELECT       0x14

/* Transmit wait select register */
#define TX_WAIT_SELECT         0x18
#define  TX_WAIT_STAT_RECV     0x40

/* Rx channel status */
#define RX_DBDMA_REG           0x40
#define RX_CHANNEL_CONTROL     0x40
#define RX_CHANNEL_STATUS      0x44
#define  RX_STATUS_RUN         1<<15
#define  RX_STATUS_PAUSE       1<<14
#define  RX_STATUS_WAKE        1<<12
#define  RX_STATUS_DEAD        1<<11
#define  RX_STATUS_ACTIVE      1<<10
#define  RX_STATUS_BT          1<<8
#define  RX_STATUS_EOP         1<<6

#define  RX_CHANNEL_RUN        RX_STATUS_RUN
#define  RX_CHANNEL_PAUSE      RX_STATUS_PAUSE
#define  RX_CHANNEL_WAKE       RX_STATUS_WAKE
#define  RX_CHANNEL_DEAD       RX_STATUS_DEAD
#define  RX_CHANNEL_ACTIVE     RX_STATUS_ACTIVE
#define  RX_CHANNEL_BT         RX_STATUS_BT
#define  RX_CHANNEL_EOP        RX_STATUS_EOP

#define  RX_DBDMA_ENABLE       (RX_CHANNEL_WAKE | RX_CHANNEL_PAUSE | \
                                RX_CHANNEL_RUN)

/*  Receive command ptr lo  */
#define RX_CMD_PTR_LO          0x4c

/*  Receive interrupt select register */
#define RX_INT_SELECT          0x50
#define  RX_INT_SELECT_EOP     0x40

/*  Receive branch select  */
#define RX_BRANCH_SELECT       0x54
#define  RX_BRANCH_SELECT_EOP  0x40

/*  Receive wait select  */
#define RX_WAIT_SELECT         0x58
#define  RX_WAIT_SELECT_EOP    0x40

/*  Event status register  */
#define EVENT_STATUS           0x80
#define  EVENT_TXSR            1<<2
#define  EVENT_EOP             1<<1
#define  EVENT_TXABORT         1<<0

/*  Interrupt enable register  */
#define INTERRUPT_ENABLE       0x82

/*  Interrupt clear register  */
#define INTERRUPT_CLEAR        0x84

/*  Interrupt status register  */
#define INTERRUPT_STATUS_REG   0x86

/*  bits for the above three interrupt registers */
#define  INTERRUPT_INTE        1<<15   /* interrupt enable */
#define  INTERRUPT_WI          1<<9    /* wakeup interrupt */
#define  INTERRUPT_ERI         1<<8    /* early recieve interrupt */
#define  INTERRUPT_PPET        1<<7    /* PCI Tx parity error */
#define  INTERRUPT_PBFT        1<<6    /* PCI Tx bus fault */
#define  INTERRUPT_IIDT        1<<5    /* illegal instruction Tx */
#define  INTERRUPT_DIT         1<<4    /* DBDMA Tx interrupt */
#define  INTERRUPT_PPER        1<<3    /* PCI Rx parity error */
#define  INTERRUPT_PBFR        1<<2    /* PCI Rx bus fault */
#define  INTERRUPT_IIDR        1<<1    /* illegal instruction Rx */
#define  INTERRUPT_DIR         1<<0    /* DBDMA Rx interrupt */

#define  INTERRUPT_TX_MASK     (INTERRUPT_PBFT|INTERRUPT_IIDT| \
                                INTERRUPT_PPET|INTERRUPT_DIT)
#define  INTERRUPT_RX_MASK     (INTERRUPT_PBFR|INTERRUPT_IIDR| \
                                INTERRUPT_PPER|INTERRUPT_DIR)

/*  chip revision register */
#define CHIP_REVISION_REG      0x8c
#define  CHIP_PCIREV_MASK      (0xf<<16)
#define  CHIP_PCIDEV_MASK      0xff

/*  Tx threshold register */
#define TX_THRESHOLD           0x94

/*  General purpose register */
#define GEN_PURPOSE_REG        0x9e

/*  General purpose pin control reg */
#define GEN_PIN_CONTROL_REG    0x9f

/*  DBDMA control register  */
#define DBDMA_CONTROL          0x90
#define  DBDMA_SRST            1<<31
#define  DBDMA_TDPCE           1<<23
#define  DBDMA_BE              1<<22
#define  DBDMA_TAP_MASK        (1<<19|1<<20|1<<21)
#define  DBDMA_RAP_MASK        (1<<16|1<<17|1<<18)
#define  DBDMA_DPMRLE          1<<15
#define  DBDMA_WIE             1<<14
#define  DBDMA_MP              1<<13
#define  DBDMA_SME             1<<12
#define  DBDMA_CME             1<<11
#define  DBDMA_DDPE            1<<10
#define  DBDMA_TDPE            1<<9
#define  DBDMA_EXTE            1<<8
#define  DBDMA_BST_MASK        (1<<4|1<<5|1<<6)
#define  DBDMA_BSR_MASK        (1<<0|1<<1|1<<2)

#define  DBDMA_BURST_1         (0x00)
#define  DBDMA_BURST_2         (0x01)
#define  DBDMA_BURST_4         (0x02)
#define  DBDMA_BURST_8         (0x03)
#define  DBDMA_BURST_16        (0x04)
#define  DBDMA_BURST_32        (0x05)
#define  DBDMA_BURST_64        (0x06)
#define  DBDMA_BURST_128       (0x07)

#define  DBDMA_TX_BST_SHIFT    (4)
#define  DBDMA_RX_BST_SHIFT    (0)

#define  DBDMA_TX_ARBITRATION_DEFAULT ( 1 << 19 )
#define  DBDMA_RX_ARBITRATION_DEFAULT ( 2 << 16 )


/*  Back-to-back interpacket gap register */
#define BTOB_INTP_GAP          0xa2
#define  BTOB_INTP_DEFAULT     0x18

/*  Non-back-to-back interpacket gap register */
#define NBTOB_INTP_GAP         0xa4

/*  MIIM command register */
#define MIIM_COMMAND           0xa6
#define  MIIM_SCAN             1<<1
#define  MIIM_RSTAT            1<<0

/*  MII address register */
#define MII_ADDRESS            0xa8
#define  MII_FIAD_MASK         (1<<8|1<<9|1<<10|1<<11|1<<12)
#define  MII_RGAD_MASK         (1<<0|1<<1|1<<2|1<<3|1<<4)

#define TPPMD_CONTROL_REG      0xa8
#define  TPPMD_FO              1<<1
#define  TPPMD_LB              1<<0

/*  MII read and write registers */
#define MII_WRITE_DATA         0xaa
#define MII_READ_DATA          0xac

/*  MII indicators */
#define MII_INDICATOR          0xae
#define  MII_NVALID            1<<2
#define  MII_SCAN              1<<1
#define  MII_BUSY              1<<0

/*  Address filter  */
#define ADDRESS_FILTER         0xd0
#define  ADDRESS_RPPRM         1<<3       /* multicast promis. mode */
#define  ADDRESS_RPPRO         1<<2       /* promiscuous mode */
#define  ADDRESS_RPAMC         1<<1       /* accept multicasts */
#define  ADDRESS_RPABC         1<<0       /* accept broadcasts */

/*  Station addresses

    Note that if the serial EEPROM is disabled, these values are all
    zero.  If, like us, you get the chips when they're fresh, they're
    also zero and you have to initialize the address */
#define STATION_ADDRESS_0      0xd2
#define STATION_ADDRESS_1      0xd4
#define STATION_ADDRESS_2      0xd6

/*  Hash tables  */
#define HASH_TABLE_0           0xd8
#define HASH_TABLE_1           0xda
#define HASH_TABLE_2           0xdc
#define HASH_TABLE_3           0xde

/* PHY indentifiers */
#define PHY_IDENTIFIER_0       0xe4
#define PHY_IDENTIFIER_1       0xe6

/*  MII Auto-negotiation register definitions  */

#define MII_AUTO_NEGOTIATION_CONTROL        (0x0000)
#define   MANC_PHY_RESET                      (0x8000)
#define   MANC_PHY_LOOPBACK_ENABLE            (0x4000)
#define   MANC_PHY_LOOPBACK_DISABLE           (0x0000)
#define   MANC_PHY_SPEED_100                  (0x2000)
#define   MANC_PHY_SPEED_10                   (0x0000)
#define   MANC_AUTO_NEGOTIATION_ENABLE        (0x1000)
#define   MANC_AUTO_NEGOTIATION_DISABLE       (0x0000)
#define   MANC_PHY_POWER_DOWN                 (0x0800)
#define   MANC_PHY_POWER_UP                   (0x0000)
#define   MANC_ISOLATE_ENABLE                 (0x0400)
#define   MANC_ISOLATE_DISABLE                (0x0000)
#define   MANC_RESTART_AUTO_NEGOTIATION       (0x0200)
#define   MANC_FULL_DUPLEX                    (0x0100)
#define   MANC_HALF_DUPLEX                    (0x0000)

#define MII_AUTO_NEGOTIATION_STATUS         (0x0001)
#define   MANS_100BASE_T4_HALF_DUPLEX         (0x8000)
#define   MANS_100BASE_X_FULL_DUPLEX          (0x4000)
#define   MANS_100BASE_X_HALF_DUPLEX          (0x2000)
#define   MANS_10MBS_FULL_DUPLEX              (0x1000)
#define   MANS_10MBS_HALF_DUPLEX              (0x0800)
#define   MANS_AUTO_NEGOTIATION_COMPLETE      (0x0020)
#define   MANS_REMOTE_FAULT                   (0x0010)
#define   MANS_AUTO_NEGOTIATION_ABILITY       (0x0008)
#define   MANS_LINK_STATUS                    (0x0004)
#define   MANS_JABBER_DETECT                  (0x0002)
#define   MANS_EXTENDED_CAPABILITY            (0x0001)

#define MII_PHY_IDENTIFIER_1                (0x0002)
#define MII_PHY_IDENTIFIER_2                (0x0003)

#define MII_AUTO_NEGOTIATION_ADVERTISEMENT  (0x0004)
#define   MANA_NEXT_PAGE                      (0x8000)
#define   MANA_REMOTE_FAULT                   (0x2000)
#define   MANA_TECHNOLOGY_ABILITY_MASK        (0x1FE0)
#define     MANATECH_10BASET_HALF_DUPLEX        (0x0020)
#define     MANATECH_10BASET_FULL_DUPLEX        (0x0040)
#define     MANATECH_100BASETX_HALF_DUPLEX      (0x0080)
#define     MANATECH_100BASETX_FULL_DUPLEX      (0x0100)
#define     MANATECH_100BASET4                  (0x0200)
#define   MANA_SELECTOR_MASK                  (0x001F)
#define     MANASELECTOR_802_3                  (0x0001)

#define MII_AUTO_NEGOTIATION_LINK_PARTNER   (0x0005)
#define   MANLP_NEXT_PAGE                     (0x8000)
#define   MANLP_ACKNOWLEDGE                   (0x4000)
#define   MANLP_REMOTE_FAULT                  (0x2000)
#define   MANLP_TECHNOLOGY_ABILITY_MASK       (0x1FE0)
#define   MANLP_SELECTOR_MASK                 (0x001F)

#define MII_AUTO_NEGOTIATION_EXPANSION      (0x0006)
#define   MANE_PARALLEL_DETECTION_FAULT       (0x0010)
#define   MANE_LINK_PARTNER_NEXT_PAGE_ABLE    (0x0008)
#define   MANE_NEXT_PAGE_ABLE                 (0x0004)
#define   MANE_PAGE_RECEIVED                  (0x0002)
#define   MANE_LINK_PARTNER_AUTO_ABLE         (0x0001)

#define MII_AUTO_NEGOTIATION_NEXT_PAGE_TRANSMIT (0x0007)
#define   MANNPT_NEXT_PAGE                    (0x8000)
#define   MANNPT_MESSAGE_PAGE                 (0x2000)
#define   MANNPT_ACKNOWLEDGE_2                (0x1000)
#define   MANNPT_TOGGLE                       (0x0800)
#define   MANNPT_MESSAGE_FIELD_MASK           (0x07FF)

#endif
