/*
 * bw_iface.h : socket abstraction layer interface.
 *
 * This file defines the external interface to the socket
 * abstraction layer defined in net/bw_sock.c
 *
 * Author: Chris Johnson <cjohnson@cobaltnet.com>
 *
 * Copyright (C) 1999, Cobalt Networks, Inc.
 *
 *	This program may be used and distributed under the
 *	terms of the GNU General Public License published by
 *	the Free Software Foundation; either version 2 of the
 *	License, or (at your option) any later version.
 *
 * Version: $Revision: 1.6 $
 */

/* 
 * Tacky CPP hacking to keep code clean.
 * Abstract differences between 2.2 and 2.0
 */
#if LINUX_VERSION_CODE >= 0x20200

#define BW_SEND_PARAMS	struct scm_cookie *scm
#define BW_SEND_OUT	scm
#define BW_RECV_PARAMS  int bw_recvflags, struct scm_cookie *scm
#define BW_RECV_OUT	bw_recvflags, scm

#else if LINUX_VERSION_CODE > 0x20000

#define BW_SEND_PARAMS	int nonblock, int bw_sendflags
#define BW_SEND_OUT	nonblock, bw_sendflags
#define BW_RECV_PARAMS	int nonblock, int bw_recvflags, int *addr_len
#define BW_RECV_OUT	nonblock, bw_recvflags, addr_len

#endif

/*
 * Interface functions to the socket managment layer. 
 * pre* functions are called before the cooresponding protocol routine, 
 * post* funstions are called after, and do* functions replace the 
 * underlying layer.
 */
struct bw_iface_funcs {
    int (*new_sock)(struct socket *, int family, int protocol);
    void (*delete_sock)(struct socket *);
    void (*setaddr)(struct socket *);

    int (*pre_dup_sock)(struct socket *newsock, struct socket *oldsock);
#if LINUX_VERSION_CODE < 0x20200
    int (*post_select)(struct socket *, int sel_type, select_table *, int pret);
#else if LINUX_VERSION_CODE >= 0x20200
    unsigned int (*post_poll)(struct file *, struct socket *,
		    struct poll_table_struct *, unsigned int pret);
#endif
    int (*post_ioctl)(struct socket *, unsigned int cmd,
		    unsigned long arg, int proto_ret);

    int (*do_sendmsg)(struct socket *, struct msghdr *, int len, BW_SEND_PARAMS);
    int (*do_recvmsg)(struct socket *, struct msghdr *, int len, BW_RECV_PARAMS);
};

int bw_sock_register_mgmt(struct bw_iface_funcs *);
int bw_sock_unregister_mgmt(void);
