/*
 * bw_sock.h : socket abstraction layer.
 *
 * This file inserts an abstraction layer between the protocol
 * and socket layers.  For now, this is used to intercept key
 * calls for the management layer.  More uses are possible.
 *
 * Author: Chris Johnson <cjohnson@cobaltnet.com>
 *
 * Copyright (C) 1999, Cobalt Networks, Inc.
 *
 *	This program may be used and distributed under the
 *	terms of the GNU General Public License published by
 *	the Free Software Foundation; either version 2 of the
 *	License, or (at your option) any later version.
 *
 * Version: $Revision: 1.10 $
 */

#include <linux/types.h>
#include <linux/uio.h>
#include <linux/time.h>
#include <linux/timer.h>
#include <linux/sched.h>
#include <linux/net.h>
#include <net/sock.h>
#include <linux/socket.h>
#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/poll.h>
#include <linux/kmod.h>
#include <linux/bw_iface.h>

/*
 * Grumble.  The "create" call doesn't get passed the family.
 *
 * This means that we need a different create for each NPROTO
 * so that we can tell where we came from.
 */

int bw_sock_docreate(struct socket *sock, int protocol, int family);

#if LINUX_VERSION_CODE < 0x20200
int real_sock_register(int fam, struct proto_ops *ops);
#else
int real_sock_register(struct net_proto_family *fam);
#endif

int real_sock_unregister(int family);

#define DPRINTF if (0) printk

#define DECL_BW_SOCK_CREATE(num)					\
int									\
bw_sock_create_##num(struct socket *sock, int protocol)			\
{									\
    return bw_sock_docreate(sock, protocol, num);			\
}

DECL_BW_SOCK_CREATE(0)
DECL_BW_SOCK_CREATE(1)
DECL_BW_SOCK_CREATE(2)
DECL_BW_SOCK_CREATE(3)
DECL_BW_SOCK_CREATE(4)
DECL_BW_SOCK_CREATE(5)
DECL_BW_SOCK_CREATE(6)
DECL_BW_SOCK_CREATE(7)
DECL_BW_SOCK_CREATE(8)
DECL_BW_SOCK_CREATE(9)
DECL_BW_SOCK_CREATE(10)
DECL_BW_SOCK_CREATE(11)
DECL_BW_SOCK_CREATE(12)
DECL_BW_SOCK_CREATE(13)
DECL_BW_SOCK_CREATE(14)
DECL_BW_SOCK_CREATE(15)
DECL_BW_SOCK_CREATE(16)
DECL_BW_SOCK_CREATE(17)
DECL_BW_SOCK_CREATE(18)
DECL_BW_SOCK_CREATE(19)
DECL_BW_SOCK_CREATE(20)
DECL_BW_SOCK_CREATE(21)
DECL_BW_SOCK_CREATE(22)
DECL_BW_SOCK_CREATE(23)
DECL_BW_SOCK_CREATE(24)
DECL_BW_SOCK_CREATE(25)
DECL_BW_SOCK_CREATE(26)
DECL_BW_SOCK_CREATE(27)
DECL_BW_SOCK_CREATE(28)
DECL_BW_SOCK_CREATE(29)
DECL_BW_SOCK_CREATE(30)
DECL_BW_SOCK_CREATE(31)

int (*bw_create_funcs[NPROTO])(struct socket *sock, int protocol) = {
    bw_sock_create_0, bw_sock_create_1, bw_sock_create_2, bw_sock_create_3,
    bw_sock_create_4, bw_sock_create_5, bw_sock_create_6, bw_sock_create_7,
    bw_sock_create_8, bw_sock_create_9, bw_sock_create_10, bw_sock_create_11,
    bw_sock_create_12, bw_sock_create_13, bw_sock_create_14, bw_sock_create_15,
    bw_sock_create_16, bw_sock_create_17, bw_sock_create_18, bw_sock_create_19,
    bw_sock_create_20, bw_sock_create_21, bw_sock_create_22, bw_sock_create_23,
    bw_sock_create_24, bw_sock_create_25, bw_sock_create_26, bw_sock_create_27,
    bw_sock_create_28, bw_sock_create_29, bw_sock_create_30, bw_sock_create_31,
};


struct proto_ops *bw_proto_reallist[NPROTO];
struct proto_ops bw_proto_mylist[NPROTO];
extern struct proto_ops bw_sock_proto;

#if LINUX_VERSION_CODE < 0x20200

int
sock_register(int fam, struct proto_ops *ops)
{
#if defined(CONFIG_COBALT_BWMGMT_DRIVER) && CONFIG_COBALT_BWMGMT_DRIVER == y
    static int bw_init = 0;

    if ( ! bw_init) {
	bw_init_all();
	bw_init = 1;
    }
#endif

    /*
     * Don't regulate pipes.  For now.
     */
    if (fam->family == AF_LOCAL) {
	return real_sock_register(fam);
    }

    if (fam >= NPROTO) {
	printk("bw sock_register called for bad protocol %d\n", fam);
	return -ENOBUFS;
    }
    /*
     * Remember who called us
     */
    bw_proto_reallist[fam] = ops;
    /*
     * Clone our entry points
     */
    bw_proto_mylist[fam] = *bw_sock_proto;
    /*
     * Set our ID
     */
    bw_proto_mylist[fam].family = fam;
    return real_sock_register(fam, &bw_sock_proto);
}

#else if LINUX_VERSION_CODE > 0x20200 

struct net_proto_family *bw_proto_families[NPROTO];
struct net_proto_family bw_proto_myfamilies[NPROTO];

/*
 * This is the hijacked register routine.  We update
 * the various tables, and forward to the real code.
 */
int
sock_register(struct net_proto_family *fam)
{
    /*
     * Don't regulate pipes.  For now. :-)
     * Don't regulate "packet" driver.  Tcpdump, etc.
     */
    if (fam->family == AF_LOCAL || fam->family == PF_PACKET) {
	return real_sock_register(fam);
    }

    if (fam->family >= NPROTO) {
	printk("bw sock_register called for bad protocol %d\n", fam->family);
	return -ENOBUFS;
    }
    DPRINTF("sock_register: family %d\n", fam->family);
    /*
     * Remember who called us
     */
    bw_proto_families[fam->family] = fam;

    /*
     * Clone the data
     */
    bw_proto_myfamilies[fam->family] = *fam;
    /*
     * Steal the entry point
     */
    bw_proto_myfamilies[fam->family].create = bw_create_funcs[fam->family];

    return real_sock_register(&bw_proto_myfamilies[fam->family]);
}

#endif

int
sock_unregister(int family)
{
    int ret;

    if ((ret = real_sock_unregister(family)) == -1)
	return ret;
    /*
     * Don't regulate pipes.  For now.
     */
    if (family == AF_LOCAL) {
	return ret;
    }

    if (family < 0 || family > NPROTO) {
	return -1;
    }


#if LINUX_VERSION_CODE > 0x20200 
    bw_proto_families[family] = NULL;
    memset(&bw_proto_myfamilies[family], 0, sizeof(bw_proto_myfamilies[family]));
#endif

    memset(&bw_proto_mylist[family], 0, sizeof(bw_proto_mylist[family]));

    return 0;
}

static struct bw_iface_funcs *bw_iface;

#include <linux/module.h>

EXPORT_SYMBOL(bw_sock_register_mgmt);
EXPORT_SYMBOL(bw_sock_unregister_mgmt);

int bw_sock_register_mgmt(struct bw_iface_funcs *funcs)
{
    bw_iface = funcs;
    return 0;
}

int bw_sock_unregister_mgmt(void)
{
    bw_iface = NULL;
    /*
     * Need to arrange for memory to be freed
     *   - possibly lazy on socket close.
     */
    return 0;
}

/*
 * In 2.0.36, bw_sock_create is in the proto_ops struct.
 * in 2.2, bw_sock_create has moved into the net_proto_family.
 */
int
bw_sock_docreate(struct socket *sock, int protocol, int family)
{
    int		(*create_func)(struct socket *sock, int protocol);
    int		ret;
    static int bw_init = 0;

    if ( ! bw_init) {
	request_module("bandwidth_mgr");
	bw_init = 1;
    }

    /*
     * This is a little tacky, but if we are out of
     * resources, so is everyone else.
     */
    DPRINTF("bw_sock_docreate: family %d\n", family);
    if (bw_iface) {
	if ((ret = bw_iface->new_sock(sock, protocol, family)) < 0) {
	    return ret;
	}
    }

#if LINUX_VERSION_CODE >= 0x20200
    /*
     * In 2.2, create lives in the proto_family structure,
     * and the protocol initializes the ops field.
     *
     * Get the create func...
     */
    create_func = bw_proto_families[family]->create;
#else if LINUX_VERSION_CODE > 0x20000
    /* 
     * In 2.0 create is called after sock_create has
     * initialized the sock->ops field.  This means
     * that we have our own routines in the socket
     * struct during the following code.
     *
     * Set the ops field and get the create func.
     */
    sock->real_ops = bw_proto_reallist[fam];

    create_func = sock->real_ops->create;
#endif

    ret = (*create_func)(sock, protocol);
    DPRINTF("create returned %d\n", ret);

    if (ret < 0) {
	if (bw_iface)
	    bw_iface->delete_sock(sock);
	return ret;
    }

#if LINUX_VERSION_CODE >= 0x20200
    /*
     * In 2.2, the protocol just initialized the ops field.
     * Steal it, and substitute our own...
     */
    sock->real_ops = sock->ops;
    sock->ops = &bw_sock_proto;
#endif

    return ret;
}

int
bw_sock_dup(struct socket *newsock, struct socket *oldsock)
{
    int ret;

#if LINUX_VERSION_CODE < 0x20200
    newsock->real_ops = oldsock->real_ops;
#endif

    if (bw_iface && (ret = bw_iface->pre_dup_sock(newsock, oldsock)) < 0)
	return ret;

    ret = oldsock->real_ops->dup(newsock, oldsock);
    DPRINTF("bw_sock_dup returning %d\n", ret);

    if (ret < 0) {
	if (bw_iface)
	    bw_iface->delete_sock(oldsock);
	return ret;
    }

#if LINUX_VERSION_CODE >= 0x20200
    newsock->real_ops = oldsock->real_ops;
    newsock->ops = &bw_sock_proto;
#endif
    return ret;
}

int
bw_sock_bind(struct socket *sock, struct sockaddr *umyaddr, 
		int sockaddr_len)
{
    int ret;

    ret = sock->real_ops->bind(sock, umyaddr, sockaddr_len);
    DPRINTF("bw_sock_bind returning %d\n", ret);

    if (ret >= 0 && bw_iface)
	bw_iface->setaddr(sock);

    return ret;
}

int
bw_sock_connect(struct socket *sock, struct sockaddr *uservaddr, 
		int sockaddr_len, int flags)
{
    int ret;
 
    ret = sock->real_ops->connect(sock, uservaddr, sockaddr_len, flags);

    if (ret >= 0 && bw_iface)
	bw_iface->setaddr(sock);

    DPRINTF("bw_sock_connect returning %d\n", ret);

    return ret;
}

int
bw_sock_socketpair(struct socket *sock1, struct socket *sock2)
{
    int ret;
 
    ret = sock1->real_ops->socketpair(sock1, sock2);
    DPRINTF("bw_sock_socketpair returning %d\n", ret);

    return ret;
}

int
bw_sock_accept(struct socket *sock, struct socket *newsock,   
		int flags)
{
    int ret;
 
    ret = sock->real_ops->accept(sock, newsock, flags);
    DPRINTF("bw_sock_accept returning %d\n", ret);

    if (ret >= 0 && bw_iface)
	bw_iface->setaddr(newsock);

    return ret;
}

int
bw_sock_getname(struct socket *sock, struct sockaddr *uaddr,   
		int *usockaddr_len, int peer)
{
    int ret;
 
    ret = sock->real_ops->getname(sock, uaddr, usockaddr_len, peer);
    DPRINTF("bw_sock_getname returning %d\n", ret);

    return ret;
}

#if LINUX_VERSION_CODE < 0x20200 
int
bw_sock_select(struct socket *sock, int sel_type, select_table *wait)
{
    int ret;
    int nio;

    ret = sock->real_ops->select(sock, sel_type, wait);
    DPRINTF("bw_sock_select got %d from lower\n", ret);

    if (sel_type == SEL_EX)
	return ret;

    if (ret >= 0 && bw_iface)
	ret = bw_iface->post_select(sock, sel_type, wait, ret);

    DPRINTF("bw_sock_select returning %d\n", ret);
    return ret;
}
#else if LINUX_VERSION_CODE >= 0x20200
unsigned int
bw_sock_poll(struct file *file, struct socket *sock,
		struct poll_table_struct *wait)
{
    int prot_ret;

    prot_ret = sock->real_ops->poll(file, sock, wait);
    DPRINTF("bw_sock_poll got %d from lower layers\n", prot_ret);

    if (bw_iface)
	prot_ret = bw_iface->post_poll(file, sock, wait, prot_ret);

    return prot_ret;
}
#endif

int
bw_sock_ioctl(struct socket *sock, unsigned int cmd, unsigned long arg)
{
    int ret;
 
    ret = sock->real_ops->ioctl(sock, cmd, arg);
    DPRINTF("bw_sock_ioctl got %d from lower\n", ret);

    if (bw_iface)
	ret = bw_iface->post_ioctl(sock, cmd, arg, ret);

    return ret;
}

int
bw_sock_listen(struct socket *sock, int len)
{
    int ret;
 
    ret = sock->real_ops->listen(sock, len);
    DPRINTF("bw_sock_listen returning %d\n", ret);

    return ret;
}

/*
 * Zap the connection.  
 */
int
bw_sock_release(struct socket *sock, struct socket *peer)
{
    int ret;
 
    ret = sock->real_ops->release(sock, peer);
    DPRINTF("bw_sock_release socket 0x%x returning %d\n", (int) sock, ret);

    if (bw_iface)
	bw_iface->delete_sock(sock);

    return ret;
}

int
bw_sock_shutdown(struct socket *sock, int flags)
{
    int ret;
 
    ret = sock->real_ops->shutdown(sock, flags);
    DPRINTF("bw_sock_shutdown returning %d\n", ret);

    return ret;
}

int
bw_sock_setsockopt(struct socket *sock, int level, int optname,   
		char *optval, int optlen)
{
    int ret;
 
    ret = sock->real_ops->setsockopt(sock, level, optname,
		optval, optlen);
    DPRINTF("bw_sock_setsockopt returning %d\n", ret);

    return ret;
}

int
bw_sock_getsockopt(struct socket *sock, int level, int optname,   
		char *optval, int *optlen)
{
    int ret;
 
    ret = sock->real_ops->getsockopt(sock, level, optname,
		    optval, optlen);
    DPRINTF("bw_sock_getsockopt returning %d\n", ret);

    return ret;
}

int
bw_sock_fcntl(struct socket *sock, unsigned int cmd,
		unsigned long arg)
{
    int ret;
 
    ret = sock->real_ops->fcntl(sock, cmd, arg);
    DPRINTF("bw_sock_fcntl returning %d\n", ret);

    return ret;
}

int
bw_sock_sendmsg(struct socket *sock, struct msghdr *m, int total_len,
		BW_SEND_PARAMS)
{
    int ret;

    if (bw_iface)
	ret = bw_iface->do_sendmsg(sock, m, total_len, BW_SEND_OUT);
    else ret = sock->real_ops->sendmsg(sock, m, total_len, BW_SEND_OUT);

    return ret;
}

int
bw_sock_recvmsg(struct socket *sock, struct msghdr *m, int total_len,
	    BW_RECV_PARAMS)
{
    int ret;

    if (bw_iface)
	ret = bw_iface->do_recvmsg(sock, m, total_len, BW_RECV_OUT);
    else ret = sock->real_ops->recvmsg(sock, m, total_len, BW_RECV_OUT);

    DPRINTF("bw_sock_recvmsg returning %d\n", ret);

    return ret;
}

struct proto_ops bw_sock_proto = {
	0,
#if LINUX_VERSION_CODE < 0x20200
	bw_sock_create,
#endif
	bw_sock_dup,
	bw_sock_release,
	bw_sock_bind,
	bw_sock_connect,
	bw_sock_socketpair,
	bw_sock_accept,
	bw_sock_getname,
#if LINUX_VERSION_CODE < 0x20200
	bw_sock_select,
#else
	bw_sock_poll,
#endif
	bw_sock_ioctl,
	bw_sock_listen,
	bw_sock_shutdown,
	bw_sock_setsockopt,
	bw_sock_getsockopt,
	bw_sock_fcntl,
	bw_sock_sendmsg,
	bw_sock_recvmsg,
};
